from dios.lib import *
from dios.options import *
import pandas as pd
import operator as op

from collections import OrderedDict
from pandas._libs.lib import to_object_array
from pandas.core.dtypes.common import (
    is_list_like,
    is_nested_list_like,
    is_scalar,
    is_integer,
    is_dict_like,
)
from pandas.core.dtypes.common import is_iterator as _is_iterator


def is_iterator(obj):
    """ This is only a dummy wrapper, to warn that the docu of this isnt't right.
    Unlike the example says,
    >>is_iterator([1, 2, 3])
    returns False, not True for lists
    """
    return _is_iterator(obj)


class DictOfSeries:
    """
    DictionaryOfSeries is a collection of pd.Series's which aim to be as close as possible similar to
    pd.DataFrame. The advantage over pd.DataFrame is, that every `column` has its own row-index,
    unlike the former, which provide a single row-index for all columns. This solves problems with
    unaligned pd.Series.

    Dataframe vs DictionaryOfSeries
    -------------------------------

    pd.DataFrame :
      - ``df['a']``      indexing with single label, work on columns, return series
      - ``df[iterabel]`` indexing with list, work on columns, return df
      - ``df[slice]``    indexing with slice, work on rows, return df
      - ``df.loc[..]``   loc[cation] based indexing, work on row and/or columns `labels`, return df
      - ``df.iloc[..]``  i[nteger]loc[cation] based indexing, work on row and/or columns, return df

    DictionaryOfSeries:
      - ``dios['a']``       indexing with single label, work on columns, return series
      - ``dios[iterable]``  indexing with list, work on columns, return dios
      - ``dios[slice]``     indexing with slice, work on (all) index'es, return dios
      - ``dios.loc[..]``    loc[cation] based indexing, work on row and/or columns `labels`, return dios
      - ``dios.iloc[..]``   i[nteger]loc[cation] based indexing, work on row and/or columns, return dios

    Todos:
    -----
    todo: to_discuss!! allow any hashable obj as column identifier
         Currently we only allow strings as identifier, to be more df-like we should allow any
         hashable object (unlike df we may should exclude stuff like: ``None`` or ``np.nan`` ??)

    """

    def __init__(self, data=None, itype=None, columns=None, downcast_policy='lossless'):

        self._data = OrderedDict()

        # We need to keep track of the index-type (itype) of every new Series.
        # If the itypes differ between different series, slicing will almost always fail
        # (eg. a datetime-like slice cannot work on a numeric index and vice versa).
        #
        # May data was given, so we firstly set itype to MixedItype, then insert all data,
        # and check/cast the itype afterwards, otherwise __setitem_new() will set the itype,
        # which may prevent inserting series with other (higher) itypes.
        self._itype = MixedItype

        policies = ['force', 'lossless', 'never']
        if downcast_policy not in policies:
            raise ValueError(f"downcast_policy must be one of {policies}")
        self._downcast_policy = downcast_policy

        if data is not None:
            self.__init_insert_data__(data)

        # we use the columns.setter to make all necessary checks
        if columns is not None:
            self.columns = columns

        # infer the itype by the data
        inferred_itype = self.__find_least_common_itype()
        itype = inferred_itype if itype is None else get_itype(itype)

        # We use the itype.setter to make all checks. If the given itype was of a lower type
        # than the inferred itype, a cast is tried on every series.
        if itype is not None:
            self.itype = itype

        # user created a empty dios: data=None(->inferred=None), itype=None
        else:
            self._itype = None

    def __init_insert_data__(self, data):
        if isinstance(data, DictOfSeries):
            for k in data:
                self[k] = data[k]

        if is_iterator(data):
            data = list(data)

        if is_dict_like(data):
            for k in data:
                self[k] = data[k]
            return

        # take care: dict's also list-like
        if is_nested_list_like(data):
            for i, d in enumerate(data):
                self[str(i)] = d
            return

        if is_list_like(data):
            self['0'] = data
            return

    def __find_least_common_itype(self):

        def all_itypes_le(itypes, super_itype):
            for itype in itypes:
                if itype_le(itype, super_itype):
                    continue
                return False
            return True

        itypes = []
        for k in self.columns:
            itypes.append(get_itype(self._data[k].index))

        if not itypes:
            return None

        found = None

        # check supertypes
        super_itypes = [MixedItype, NumericItype]
        for super_itype in super_itypes:
            if all_itypes_le(itypes, super_itype):
                found = super_itype
                continue
            break
        assert found, "At least this should be MixedItype"

        # check base types
        single_itypes = [DatetimeItype, IntegerItype, FloatItype]
        for single_itype in single_itypes:
            if all_itypes_le(itypes, single_itype):
                found = single_itype
                break
        return found

    @property
    def columns(self):
        return list(self._data.keys())

    @columns.setter
    def columns(self, new):
        if not isinstance(new, list):
            raise TypeError("column names must be given as a list")

        if len(set(new)) != len(new):
            raise ValueError("column names must be unique")

        if len(new) != len(self.columns):
            raise ValueError(f"Length mismatch: Columns has {len(self.columns)} elements, "
                             f"new values have {len(new)} elements")

        # to keep order, we iterate over self instead of new
        _d = OrderedDict()
        for i, k in enumerate(self.columns):
            _d[new[i]] = self[k]
        self._data = _d

    @property
    def values(self):
        return to_object_array(self._data.values()).transpose()

    @property
    def data(self):
        return self._data.values()

    @property
    def itype(self):
        return self._itype

    @itype.setter
    def itype(self, itype_like):
        itype = get_itype(itype_like)

        if not is_itype_subtype(self._itype, itype):
            # try to cast all series to the new itype
            self.__cast_all(itype)

        self._itype = itype

        if not itype.unique:
            throwMixedItypeErrWarn(f"Using a {itype} as dios.itype is experimental. As soon as series with "
                                   f"different index types are inserted, slicing will almost always fail. "
                                   f"You are hereby warned!")

    def __cast_all(self, itype):
        for k in self.columns:
            casted = cast_to_fit_itype(self._data[k].copy(), itype, policy=self._downcast_policy)
            if casted is None:
                raise ItypeCastError(f"Cast series indicees to the given itype failed for series in column {k}.")
            self._data[k] = casted

    def _check_keys(self, keys):
        missing = [k for k in keys if k not in self.columns]
        if missing:
            raise KeyError(f"{missing} not in index")

    def __getitem__(self, key):
        """
        Get items:
          - ``dios['a']``      -> return a pd.Series
          - ``dios[iterable]`` -> return a DictOfSeries of all given columns [1]
          - ``dios[slice]``    -> return a sliced DictOfSeries of all(!) columns in the dios

        Notes:
          - [1] If ``iterable`` contains any(!) label that does not exist, a KeyError is raised.
        """

        # prepare
        if is_iterator(key):
            key = list(key)

        if isinstance(key, str):
            self._check_keys([key])
            return self._get_item(key)

        if isinstance(key, slice):
            return self._slice(self.columns, key)

        if is_list_like(key):
            self._check_keys(key)
            return self._getitem_listlike(key)

        raise KeyError(f"{key}")

    def _get_item(self, key):
        # return always a pd.Series
        return self._data[key]

    def _getitem_listlike(self, keys):
        new = self.copy_empty()
        for k in keys:
            new[k] = self._get_item(k)
        return new

    def _slice(self, keys, slicer):
        """ Return a slice of self"""
        new = self.copy_empty()
        for k in keys:
            new[k] = self._get_item(k)[slicer]
        return new

    def __setitem__(self, key, value):
        """
        Set items:
          - ``dios['a'] = pd.Series()`` -> Set a new pd.Series to an existing column or add it as `new column`.
          - ``dios['a'] = Any`` -> Pass ``any`` to the pd.Series in the corresponding existing(!) column. [1],[2]
          - ``dios[iterable] = Any`` -> Pass ``any`` to the pd.Series's in the corresponding set of columns. [1],[2],[3]
          - ``dios[slice] = Any`` -> Pass ``any`` to all(!) sliced pd.Series's from the current dios. [1],[2]

        Notes:
          - [1] The length of ``Any`` must be be equal to the length of every sliced pd.Series.
          - [2] If ``Any`` is a ``DictOfSeries`` the behavior depends on the option ``dios_to_dios_method``
            in the ``options`` dictionary.
          - [3] If ``iterable`` contains any(!) label that does not exist, a KeyError is raised.
        """

        # prepare
        if is_iterator(key):
            key = list(key)

        # determine action by keys
        if isinstance(key, str):
            if key not in self.columns:
                self._setitem_new(key, value)
            else:
                self._setitem(key, value)
            return

        elif isinstance(key, slice):
            keys = self.columns
            slicer = key

        elif is_list_like(key):
            self._check_keys(key)
            keys = key
            slicer = None

        else:
            raise KeyError(f"{key}")

        if isinstance(value, DictOfSeries):
            self._setitem_dios(keys, slicer, value)
        else:
            if is_iterator(value):
                value = list(value)

            for k in keys:
                self._setitem(k, value, slicer)

    def _setitem_new(self, key, value, bypass_checks=False):
        v = value

        # if the checks was already done, we skip them here,
        # also the Errormessage wouldn't fully apply.
        if not bypass_checks:
            if isinstance(v, DictOfSeries):
                v = v.squeeze()

            elif is_list_like(v):
                v = pd.Series(v)  # upcast

            if not isinstance(v, pd.Series):
                raise ValueError(f"Only pd.Series and DictOfSeries (of length 1) can be assigned new")

        itype = get_itype(v.index)

        if self._itype is None:
            # if the user created a empty dios
            self._itype = itype

        v = cast_to_fit_itype(v, self._itype, policy=self._downcast_policy)
        if v is None:
            raise ValueError(f"Itype mismach. Policy `{self._downcast_policy}` forbid to down-cast"
                             f"itype `{itype}` to itype `{self.itype}`. key: {key}")

        self._data[key] = v.copy(deep=True)

    def _setitem(self, key, val, sl=None):
        """ Set a value or a set of values to a single(!) key in self k"""

        # series, dios['a'] = series
        if isinstance(val, pd.Series) and sl is None:
            self._setitem_new(key, val, bypass_checks=True)
            return

        sl = sl or slice(None)
        item = self._data[key]

        # label, scalar: dios['a'] = 3.9 or
        # slice, scalar: dios[0:3] = 4.0
        if is_scalar(val):
            item[sl] = val

        # label,  list: dios['a'] = [0.0, 0.3, 0.0]
        # sclice, list: dios[0:3] = [0.0, 0.3, 0.0]
        elif is_list_like(val):
            # ensure same size # fixme: is this neccessary, wouldnt pd.Series raise a Valuerror ?
            if len(item[sl]) == len(val):
                item[sl] = val
            else:
                raise ValueError(f'Length of values does not match length of index for the key {key}')
        else:
            raise ValueError(f"assignments with a values of type {type(val)} are not supported")
        return

    def _setitem_dios(self, keys, slicer, other):
        keys = get_dios_to_dios_keys(keys, other)
        for k in keys:
            self._setitem(k, other[k], slicer)

    @property
    def loc(self):
        return _LocIndexer(self)

    @property
    def iloc(self):
        return _iLocIndexer(self)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        pd_max_rows = pd.get_option('display.max_rows')
        pd.set_option('display.max_rows', dios_options[Options.disp_max_rows])

        def item2str(k):
            kstr = str(self[k]).replace('\n', '\n ')
            return f'{k}:\n {kstr}\n'

        maxrows = dios_options[Options.disp_max_vars]
        s = ''
        head = maxrows // 2
        tail = len(self.columns) - head
        for i, k in enumerate(self.columns):
            if i < head:
                s += item2str(k)
            elif i == head:
                s += ' ... \t\t\t\t ...\n' * 2
            elif i > tail:
                s += item2str(k)
        pd.set_option('display.max_rows', pd_max_rows)
        return s

    def __bool__(self):
        raise ValueError("The truth value of a DictionaryOfSeries is ambiguous. Use a.empty, a.any() or a.all().")

    @property
    def empty(self):
        return all(self._data[c].empty for c in self._data)

    def all(self):
        return all(all(self._data[c]) for c in self._data)

    def any(self):
        return any(any(self._data[c]) for c in self._data)

    def __len__(self):
        return len(self._data)

    def __reversed__(self):
        return reversed(self._data)

    def __contains__(self, item):
        return item in self._data

    def __iter__(self):
        yield from self._data

    def __delitem__(self, key):
        del self._data[key]

    def __copy__(self):
        return self.copy(deep=True)

    def __deepcopy__(self, memo=None):
        return self.copy(deep=True)

    def copy(self, deep=True):
        new = self.copy_empty()
        # We use `_data` here, because all checks are already done.
        # So this should be much faster, especially, because we use the underlying dict for
        # getting and setting the values, instead of ``__setitem__`` and ``__getitem__``.
        # Note: don't use same approach elsewhere, unless you're very sure what you do.
        for k in self._data:
            new._data[k] = self._data[k].copy(deep=deep)
        return new

    def copy_empty(self):
        new = DictOfSeries()
        new._itype = self.itype
        return new

    def __op1__(self, op):
        new = self.copy_empty()
        for k in self.columns:
            new[k] = op(self[k])
        return new

    def __op2__(self, other, op):
        new = self.copy_empty()
        if isinstance(other, DictOfSeries):
            keys = get_dios_to_dios_keys(self.columns, other)
            for k in keys:
                new[k] = op(self[k], other[k])
        else:
            for k in self.columns:
                new[k] = op(self[k], other)
        return new

    def __neg__(self):
        return self.__op1__(op.neg)

    def __invert__(self):
        return self.__op1__(op.inv)

    def __abs__(self):
        return self.__op1__(op.abs)

    def __lt__(self, other):
        return self.__op2__(other, op.lt)

    def __le__(self, other):
        return self.__op2__(other, op.le)

    def __eq__(self, other):
        return self.__op2__(other, op.eq)

    def __ne__(self, other):
        return self.__op2__(other, op.ne)

    def __ge__(self, other):
        return self.__op2__(other, op.ge)

    def __gt__(self, other):
        return self.__op2__(other, op.gt)

    def __add__(self, other):
        return self.__op2__(other, op.add)

    def __sub__(self, other):
        return self.__op2__(other, op.sub)

    def __mul__(self, other):
        return self.__op2__(other, op.mul)

    def __mod__(self, other):
        return self.__op2__(other, op.mod)

    def __truediv__(self, other):
        return self.__op2__(other, op.truediv)

    def __floordiv__(self, other):
        return self.__op2__(other, op.floordiv)

    def __pow__(self, other):
        return self.__op2__(other, op.pow)

    def __and__(self, other):
        return self.__op2__(other, op.and_)

    def __or__(self, other):
        return self.__op2__(other, op.or_)

    def __xor__(self, other):
        return self.__op2__(other, op.xor)

    def squeeze(self):
        if len(self) == 1:
            return self[self.columns[0]]
        else:
            return self

    def memory_usage(self, index=True, deep=False):
        mem = 0
        for k in self.columns:
            mem += self[k].memory_usage(index=index, deep=deep)
        return mem

    def foreach(self, f, *args, **kwargs):
        self.pipe(f, *args, **kwargs)

    def pipe(self, f, *args, **kwargs):
        """
        Call a pd.Series function on all series in the DictOfSeries

        :param f: pd.Series function to apply on all Series
        :param args: args for the function
        :param kwargs: kwargs for the function
        :return:
        """
        news = pd.Series()
        newd = self.copy_empty()
        need_dios = False
        # return a copy nevertheless, but also run inplace if inplace=True or
        # if the function not has this option, but work inplace.

        inplace = kwargs.get('inplace', False) or kwargs.get('copy', False)

        def getobj(k, inplace):
            if inplace:
                # self[k] may return a copy, so we need to get
                # the item from the dict, by our mother class.
                return self._data[k]
            else:
                return self._data[k].copy()

        for k in self.columns:
            v = f(getobj(k, inplace), *args, **kwargs)
            if inplace:
                continue
            if isinstance(v, pd.Series):
                newd[k] = v
                need_dios = True
            else:
                newd[k] = pd.Series(data=v)
                news[k] = v

        if inplace:
            return None
        if need_dios:
            return newd.squeeze()
        if len(news) == 0:
            return None
        return news.squeeze()


class _Indexer:
    def __init__(self, _dios):
        self._dios = _dios
        # short handles
        self._data = _dios._data


class _LocIndexer(_Indexer):

    def __init__(self, _dios):
        super().__init__(_dios)
        self._check_keys = _dios._check_keys

    def __getitem__(self, key):
        rkey, cols = self._unpack_key(key)
        new = self._dios.copy_empty()
        for c in cols:
            new[c] = self._data[c].loc[rkey]
        return new

    def __setitem__(self, key, value):
        rkey, cols = self._unpack_key(key)
        # todo: dios -> dios_to_dios, -> series
        #       scalar, -> automatically
        #       series, -> automatically
        #       list_like -> check length
        for c in cols:
            self._data[c].loc[rkey] = value
        # todo loc.__setitem__(self, key, value):
        raise NotImplementedError

    def _unpack_key(self, key):
        # if we have a tuple, we have a rows- and a column-indexer
        # if not, we only have a row-indexer and work on all columns
        if isinstance(key, tuple):
            rkey, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")

            # prepare ckey
            if is_iterator(ckey):
                ckey = list(ckey)

            # determine columns
            if isinstance(ckey, str):
                self._check_keys([ckey])
                cols = [ckey]
            elif isinstance(ckey, slice):
                cols = self._col_slice_to_col_list(ckey)
            elif is_list_like(ckey):
                self._check_keys(ckey)
                cols = ckey
            else:
                raise KeyError(f"Type {type(ckey)} is not supported to select columns.")
        else:
            cols = self._data.keys()
            rkey = key
        return rkey, cols

    def _col_slice_to_col_list(self, cslice):
        """ see here:
        https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#indexing-slicing-with-labels
        """
        keys = list(self._data.keys())
        try:
            start = keys.index(cslice.start) if cslice.start is not None else None
            stop = keys.index(cslice.stop) if cslice.stop is not None else None
        except ValueError:
            raise KeyError("The slice start label or the slice stop label is not present in the columns.")
        if not is_integer(cslice.step) and cslice.step > 0:
            raise TypeError("The step parameter of the slice must be positive integer.")
        return keys[slice(start, stop + 1, cslice.step)]


class _iLocIndexer(_Indexer):

    def __getitem__(self, key):
        rkey, cols = self._unpack_key(key)
        new = self._dios.copy_empty()
        for c in cols:
            new[c] = self._data[c].iloc[rkey]
        return new

    def __setitem__(self, key, value):
        # todo iloc.__setitem__(self, key, value):
        raise NotImplementedError

    def _unpack_key(self, key):
        # if we have a tuple, we have a rows- and a column-indexer
        # if not, we only have a row-indexer and work on all columns
        if isinstance(key, tuple):
            rkey, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")

            # prepare ckey
            if is_iterator(ckey):
                ckey = list(ckey)

            # determine columns
            if is_integer(ckey):
                self._check_keys([ckey])
                cols = self._integers_to_col_list([ckey])
            elif isinstance(ckey, slice):
                cols = self._col_slice_to_col_list(ckey)
            elif is_list_like(ckey):
                self._check_keys(ckey)
                cols = self._integers_to_col_list(ckey)
            else:
                raise KeyError(f"Type {type(ckey)} is not supported for indexing on columns.")
        else:
            cols = self._data.keys()
            rkey = key
        return rkey, cols

    def _check_keys(self, keys):
        bound = len(self._data)
        for k in keys:
            if k not in range(-bound, bound):
                raise KeyError("positional indexer(s) are out-of-bounds in columns")

    def _integers_to_col_list(self, ints):
        klist = list(self._data.keys())
        ks = set()
        for i in ints:
            ks.add(klist[i])
        return ks

    def _col_slice_to_col_list(self, sl):
        for s in [sl.start, sl.stop, sl.step]:
            if not is_integer(s):
                raise TypeError(f"positional indexing with slice must be integers, passed was {s} of {type(s)}")
        return list(self._data.keys())[sl]
